// src/Terminal.hh
// This file is part of SlugTerm; see http://chezphil.org/slugterm
// (C) 2005 Philip Endecott
// Modifications for 4D Systems Display Graphics Modules (C) 2008 Sven Killig <sven@killig.de>

// This is version $Name:  $
//   (if there is no version (e.g. V0-1) mentioned in the previous line,
//    this is probably a snapshot from between "official" releases.)

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifndef Terminal_hh
#define Terminal_hh

extern "C" {
#include <rote/rote.h>
}

#include <string>
#include <boost/function.hpp>
#include <boost/thread/thread.hpp>
#include <boost/thread/mutex.hpp>


class Terminal {

private:
  RoteTerm* rt;
  typedef boost::mutex mutex_t;
  mutex_t mutex;
  typedef mutex_t::scoped_lock read_lock_t;
  typedef mutex_t::scoped_lock write_lock_t;
  boost::function<void(void)> onChange;

public:
  Terminal(int rows,
	   int cols,
	   std::string command,
	   boost::function<void(void)> onchange_);

  ~Terminal();

  int rows(void) const;
  int cols(void) const;
  RoteCell cell(int r, int c) const;
  int cur_row(void) const;
  int cur_col(void) const;
  bool curpos_dirty(void) const;
  bool line_dirty(int line) const;
  void line_dirty_set(int line, bool dirty) const;

  void write(std::string data);
  void run(void);

};


#endif
