// src/UsbKeyboard.cc
// This file is part of SlugTerm; see http://chezphil.org/slugterm
// (C) 2005 Philip Endecott
// Modifications for 4D Systems Display Graphics Modules (C) 2008 Sven Killig <sven@killig.de>

// This is version $Name:  $
//   (if there is no version (e.g. V0-1) mentioned in the previous line,
//    this is probably a snapshot from between "official" releases.)

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#include "UsbKeyboard.hh"

#include "KeyboardLayout.hh"

#include "Exception.hh"

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <linux/input.h>

#include <string>
#include <algorithm>

#include <boost/function.hpp>


using namespace std;


UsbKeyboard::UsbKeyboard(string event_dev_fn,
			 KeyboardLayout& keyboard_layout_,
			 boost::function<void(string)> onKeypress_):
  keyboard_layout(keyboard_layout_),
  onKeypress(onKeypress_)
{
  event_dev_fd = open(event_dev_fn.c_str(), O_RDONLY);
  if (event_dev_fd<0) {
///    throw SysException("open("+event_dev_fn+")");
  }
}


UsbKeyboard::~UsbKeyboard()
{
  int rc = close(event_dev_fd);
  if (rc==-1) {
///    throw SysException("close(event_dev)");
  }
}


void UsbKeyboard::run(void)
{
  KeyboardLayout::modifiers_t modifiers=0;
  while (1) {
    struct input_event ev;
    size_t read_bytes = read(event_dev_fd, &ev, sizeof(ev));
    if (static_cast<int>(read_bytes) == -1) {
///      throw SysException("read(event_dev)");
    }
    if (read_bytes < sizeof(struct input_event)) {
///      throw StrException("read(event_dev) short read");
    }

    if (ev.type!=EV_KEY) {
      continue;
    }
    bool press = (ev.value!=0);
    keyboard_layout.update_modifiers(modifiers,ev.code,press);
    if (!press) {
      continue;
    }

    string keys = keyboard_layout.decode(ev.code,modifiers);
    onKeypress(keys);
  }
}
