// src/UsbLcd.cc
// This file is part of SlugTerm; see http://chezphil.org/slugterm
// (C) 2005 Philip Endecott
// Modifications for 4D Systems Display Graphics Modules (C) 2008 Sven Killig <sven@killig.de>

// This is version $Name:  $
//   (if there is no version (e.g. V0-1) mentioned in the previous line,
//    this is probably a snapshot from between "official" releases.)

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

// apt-get install libboost-dev
// http://rote.sourceforge.net
// package/
//  boost
// cp -p ../../lcd4linux/drv_D4D.c ../../lcd4linux/drv_generic_serial.c ../../lcd4linux/drv_generic_graphic.c ../../lcd4linux/*x*_*_*SB_?.h build_mipsel/lcd4linux-r877;WGET_OPTIONS=--no-check-certificate make package/lcd4linux-compile V=99;scp build_mipsel/lcd4linux-r877/lcd4linux root@192.168.30.4:

#include "UsbLcd.hh"

#include "dldemo.h"

#include <stdint.h>
#include <unistd.h>

#include <string>
#include <vector>

#include <boost/assign/std/vector.hpp>

using namespace std;
using namespace boost::assign;

/*uint8_t dl_register_init_U70[]  = { 0x00, 0x20, 0x3c, 0x7a, 0xc9, 0xf2, 0x6c, 0x48, 0xf9, 0x70, 0x53, 0xff, 0xff, 0x21, 0x27, 0x03, 0x20, 0x91, 0xf3, 0xff, 0xff, 0xff, 0xf9, 0x01, 0xe0, 0x01, 0x02, 0xc8, 0x19 };
DisplayLink dl(0x17e9, 0x0103, dl_register_init_U70);*/
DisplayLink dl;
var_screeninfo	myscreen;
static unsigned char colortable[256][3];
std::string fpath; unsigned int fwidth; unsigned int fheight;

UsbLcd::UsbLcd(unsigned int pid, std::string init, std::string fpath, unsigned int fwidth, unsigned int fheight)
{

	unsigned char key[29];
	for (int i = 0; i < 29; i++) {
		const char tmp[3] = { init.at(2*i), init.at(2*i+1), 0 };
		if ( sscanf( tmp, "%hhx", &(key[i]) ) == 1 ) continue;
		printf("error in hex string\n"); exit(2);
	}
	dl.init(pid, key, fpath, fwidth, fheight);


// the 6 value iterations in the xterm color cube
static const unsigned char valuerange[] = { 0x00, 0x5F, 0x87, 0xAF, 0xD7, 0xFF };

// 16 basic colors
static const unsigned char basic16[16][3] =
{
	{ 0x00, 0x00, 0x00 }, // 0
	{ 0xCD, 0x00, 0x00 }, // 1
	{ 0x00, 0xCD, 0x00 }, // 2
	{ 0xCD, 0xCD, 0x00 }, // 3
	{ 0x00, 0x00, 0xEE }, // 4
	{ 0xCD, 0x00, 0xCD }, // 5
	{ 0x00, 0xCD, 0xCD }, // 6
	{ 0xE5, 0xE5, 0xE5 }, // 7
	{ 0x7F, 0x7F, 0x7F }, // 8
	{ 0xFF, 0x00, 0x00 }, // 9
	{ 0x00, 0xFF, 0x00 }, // 10
	{ 0xFF, 0xFF, 0x00 }, // 11
	{ 0x5C, 0x5C, 0xFF }, // 12
	{ 0xFF, 0x00, 0xFF }, // 13
	{ 0x00, 0xFF, 0xFF }, // 14
	{ 0xFF, 0xFF, 0xFF }  // 15
};

	unsigned short int c;
	for(c=0;c<=255;c++)
	{
	  // 16 basic colors
	  if(c<16)
	  {
		  colortable[c][0] = basic16[c][0];
		  colortable[c][1] = basic16[c][1];
		  colortable[c][2] = basic16[c][2];
	  }

	  // color cube color
	  if(c>=16 && c<=232)
	  {
		  colortable[c][0] = valuerange[((c-16)/36)%6];
		  colortable[c][1] = valuerange[((c-16)/6)%6];
		  colortable[c][2] = valuerange[(c-16)%6];
	  }

	  // gray tone
	  if(c>=233 && c<=255)
	  {
		  colortable[c][0]=colortable[c][1]=colortable[c][2] = 8+(c-232)*0x0a;
	  }
	}



backlight(true);
  myscreen=dl.getScreenInfo();
//fprintf(stderr,"%s", myscreen.id);
fprintf(stderr,"%ix%i\n", cols(), rows());
//fprintf(stderr,"font=%i\n", font);
  clear();

  set_mode(true,true,false);
}


UsbLcd::~UsbLcd()
{
}


void UsbLcd::clear(void)
{
///  uoled128gmd1.eraseScreen();
  dl.fill(0,0,myscreen.xres,myscreen.yres,0);
  current_row=0;
  current_col=0;
  current_pos_known=true;
}


void UsbLcd::home(void)
{
  current_row=0;
  current_col=0;
  current_pos_known=true;
}


bool display_,line_cursor_;
void UsbLcd::set_mode(bool enable_display,
		      bool enable_line_cursor,
		      bool enable_block_cursor)
{
//fprintf(stderr,"set_mode(%i, %i, %i)\n", enable_display, enable_line_cursor, enable_block_cursor);
  /*send(ir,
       0x8 | (enable_display?0x4:0) | (enable_line_cursor?0x2:0)
       | (enable_block_cursor?0x1:0) );*/
  /*if(display_!=enable_display) {
    display_=enable_display;
    // happens in backlight():
    if(enable_display) uoled128gmd1.powerUp();
    else uoled128gmd1.moddown();
  }*/
  if(line_cursor_!=enable_line_cursor) {
    line_cursor_=enable_line_cursor;
    int color;
    if(enable_line_cursor) color=65535;
    else color=0;
///    uoled128gmd1.drawLine(current_col*XRES,(current_row+1)*YRES-1,(current_col+1)*XRES-1,(current_row+1)*YRES-1,color);
    dl.lineHor(current_col*myscreen.xres_font,(current_row+1)*myscreen.yres_font-1,myscreen.xres_font,color);
    dl.update();
  }
  //uoled128gmd1.setDisplay(enable_display);
}


void UsbLcd::set_pos(int row, int col)
{
//fprintf(stderr,"set_pos(%i, %i)\n", row, col);
  if (current_pos_known
      && (current_row==row)
      && (current_col==col)) {
    return;
  }
  current_row=row;
  current_col=col;
  current_pos_known=true;
}


void UsbLcd::write(string s, unsigned int attr)
{
//std::cout << s;
  unsigned char r=0,g=0,b=0,i;
  unsigned char gc[2];
  unsigned short int gc16[2];
  int BG=0, FG=1;
  gc[BG]=ROTE_ATTR_BG(attr);
  gc[FG]=ROTE_ATTR_FG(attr);
  if(gc[FG]<8 && ROTE_ATTR_BOLD(attr)/* || ROTE_ATTR_BLINK(attr)*/) gc[FG]+=8;
  if(gc[BG]<8 && ROTE_ATTR_BLINK(attr)) gc[BG]+=8;
  for (i = BG; i <= FG; i++) {
      /*switch(gc[i]) {
        case 0: r= 0; g= 0; b= 0; break; // black
        case 1: r=31; g= 0; b= 0; break; // red
        case 2: r= 0; g=63; b= 0; break; // green
        case 3: r=31; g=63; b= 0; break; // yellow
        case 4: r= 0; g= 0; b=31; break; // blue
        case 5: r=31; g= 0; b=31; break; // magenta
        case 6: r= 0; g=63; b=31; break; // cyan
        case 7: r=31; g=63; b=31; break; // white
      }*/
      r=colortable[gc[i]][0];
      g=colortable[gc[i]][1];
      b=colortable[gc[i]][2];
      //if(!i || (i && !(ROTE_ATTR_BOLD(attr)/* || ROTE_ATTR_BLINK(attr)*/))) {
        /*r>>=1;
        g>>=1;
        b>>=1;
      }*/
      //gc16[i]=RGB_to16(r, g, b);
      gc16[i]=RGB_24to16(r, g, b);
  }
  //dl.render(current_row, current_col, gc16[1], gc16[0], s);
  if(ROTE_ATTR_REVERSE(attr)) {
    unsigned short int temp=gc16[BG];
    gc16[BG]=gc16[FG];
    gc16[FG]=temp;
  }
  dl.render(current_row, current_col, gc16[FG], gc16[BG], s);
  //if(ROTE_ATTR_BLINK(attr)) {
  if(ROTE_ATTR_UNDERLINE(attr)) {
///    uoled128gmd1.drawLine(current_col*XRES,(current_row+1)*YRES-1,(current_col+s.length())*XRES-1,(current_row+1)*YRES-1,gc16[1]);
    dl.lineHor(current_col*myscreen.xres_font,(current_row+1)*myscreen.yres_font-1,s.length()*myscreen.xres_font,gc16[FG]);
  }
  current_col+=s.length();
  if (current_col>=cols()) {
    current_pos_known=false;
  }
}

void UsbLcd::write(string s) {
  write(s,ROTE_DEFAULT_ATTR);
}


void UsbLcd::write_at(int row, int col, string s, unsigned int attr)
{
  set_pos(row,col);
  write(s,attr);
}

void UsbLcd::write_at(int row, int col, string s)
{
  write_at(row,col,s,ROTE_DEFAULT_ATTR);
}


void UsbLcd::backlight(bool enable)
{
//fprintf(stderr,"backlight(%i)\n", enable);
  if (backlight_!=enable) {
    backlight_=enable;
    //ftdi.write(encode(backlight_,dr,false,false,0));
    if(enable) {
      ///uoled128gmd1.powerUp();
      ///sleep(1);
      ///uoled128gmd1.setContrast(15);
    } ///else uoled128gmd1.powerDown();
  }
}


int UsbLcd::rows(void) const
{
  return myscreen.yres/myscreen.yres_font;
}


int UsbLcd::cols(void) const
{
  return myscreen.xres/myscreen.xres_font;
}


void UsbLcd::update(void)
{
  dl.update();
}
