/*
2009 by Sven Killig <sven@killig.de>
FON2FNT: http://www.gadgetparadise.com/901301/index.htm
*/

#ifndef _dldemo_H
#define _dldemo_H

#include <usb.h>
#include <unistd.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <asm/types.h>

#include <string>
//#include "8x12_horizontal_MSB_1.h"

#include <ft2build.h>
#include FT_FREETYPE_H
#include FT_WINFONTS_H
#include FT_MODULE_H

//for 2 bytes long variables
#define msb(variable) (variable >> 8)
#define lsb(variable) (variable & 0xFF)

//Copied from linux/fb.h
struct var_screeninfo {
	std::string id;			/* identification string eg "TT Builtin" */
	__u32 xres;			/* visible resolution		*/
	__u32 yres;
	__u32 xres_virtual;		/* virtual resolution		*/
	__u32 yres_virtual;
	__u32 xoffset;			/* offset from virtual to visible */
	__u32 yoffset;			/* resolution			*/
	__u32 bits_per_pixel;		/* guess what			*/
	__u32 height;			/* height of picture in mm    */
	__u32 width;			/* width of picture in mm     */
	__u32 baud_rate;		/*added */
	__u32 timeout;			/*time to wait for ACK*/
	bool writeWaitForACK;		/*added, 1 = Wait ACK */
	const char * deviceName;	/*added, name of the communication port */
	__u32 hardwareVersion;
	__u32 firmwareVersion;
	__u32 xres_font;
	__u32 yres_font;
};

/******************* COMMAND BUFFER ***********************/
typedef struct {
	uint8_t* buffer;
	int pos, size;
} dl_cmdstream;

typedef struct {
	uint8_t count;
	uint16_t value;
} rle_word;

/******************* COMMANDS *****************************/
#define DL_SOC 0xAF // start-of-command token

/******************* REGISTER COMMANDS ********************/
#define DL_REG_COLORDEPTH   0x00 // 0 = 16 bit, 1 = 24 bit
// 0x01 - 0x0E unknown
#define DL_REG_XRES_MSB     0x0F
#define DL_REG_XRES_LSB     0x10
// 0x11 - 0x16 unknown
#define DL_REG_YRES_MSB     0x17
#define DL_REG_YRES_LSB     0x18
// 0x19 - 0x1C unknown
// 0x1D - 0x1E unused
#define DL_REG_BLANK_SCREEN 0x1F // 0 = normal operation, 1 = blank screen
// 0x20 - 0xFE unused
#define DL_REG_SYNC         0xFF // 0 = hold register updates, 0xFF = resume
// The unknown registers very likely contain pixel clock, sync polarity etc.

/****************** ADDRESS REGISTERS *********************/
#define DL_ADDR_FB16_START  0x20 // 16-bit mode, color MSBs, RGB 565
#define DL_ADDR_FB16_STRIDE 0x23
#define DL_ADDR_FB8_START   0x26 // additional 8 bit for 24-bit mode, color LSBs, RGB 323?
#define DL_ADDR_FB8_STRIDE  0x29

/******************* GRAPHICS COMMANDS ********************/
#define DL_GFX_BASE 0x60 // base graphics command
#define DL_GFX_WORD 0x08 // word-mode flag
#define DL_GFX_COMP 0x10 // compressed-mode flag
#define DL_GFX_WRITE (DL_GFX_BASE | 0x00) // write memory
#define DL_GFX_RLE   (DL_GFX_BASE | 0x01) // write RLE-encoded data
#define DL_GFX_COPY  (DL_GFX_BASE | 0x02) // internal copy


//generic class to manage DisplayLink adapters
class DisplayLink
{
	public:
		DisplayLink(void);
		void init(unsigned int pid, uint8_t* init, std::string fpath, unsigned int fwidth, unsigned int fheight);
		~DisplayLink();
		var_screeninfo getScreenInfo(void);			//returns infos like screen width, height etc...
		//Generic methods
		void render(const unsigned short int row, const unsigned short int col, unsigned int fg, unsigned int bg, std::string txt);
		void render2(const unsigned short int row, const unsigned short int col, unsigned int fg, unsigned int bg, std::string txt);
		void update(void);
		void lineHor(unsigned int x, unsigned int y, unsigned int length, unsigned int colour);
		void fill(unsigned int x, unsigned int y, unsigned int width, unsigned int height, unsigned int colour);
	private:


/******************* ENCRYPTION STUFF **********************/
uint8_t  keybuffer[0x11000];
uint16_t ofsbuffer[0x1000];

// from main()
	dl_cmdstream cs;
	usb_dev_handle* handle;
uint8_t* drv_generic_graphic_FB;
  FT_Library  library;   /* handle to library     */
  FT_Face     face;      /* handle to face object */

		var_screeninfo screeninfo;


int dl_crc12( unsigned char* data, int len);
void dl_generate_key( uint8_t key[0x11000], uint16_t map[0x1000] );

uint8_t dl_peek( int addr, int timeout );
void dl_poke( int addr, uint8_t value, int timeout );
int dl_poll( int timeout );
void dl_set_key( uint8_t key[16], int timeout );
void dl_blink( int timeout );
void dl_get_edid( uint8_t edid[128], int timeout );

inline void create( int size );
inline void destroy();
inline void send( int ep, int timeout );
inline void insertb( uint8_t val );
inline void insertw( uint16_t val );
inline void inserta( uint32_t address );
inline void insertd( uint32_t val );
inline void insert( int size, uint8_t* buf );

void dl_unknown_command();
void dl_sync_command();
void dl_decomp_table_command( int size, uint8_t* buf );

void dl_set_register( uint8_t reg, uint8_t val );
void dl_set_registers( uint8_t values[0x1D] );

void dl_set_address( uint8_t reg, int address );
void dl_set_offsets( int start16, int stride16, int start8, int stride8 );
void dl_gfx_command( uint8_t cmd, int addr, uint8_t count );
void dl_gfx_write( int addr, uint8_t count, uint8_t* data );
void dl_gfx_rle( int addr, uint8_t count, rle_word* rs );
void dl_gfx_copy( int src_addr, int dst_addr, uint8_t count );

void dl_init();
void dl_dumpmem( char* f );

usb_dev_handle* usb_get_device_handle( int vendor, int product, int interface );
void rgb24_to_rgb16( uint8_t* rgb24, uint8_t* rgb16, int count );
void read_rgb24( const char* filename, uint8_t* rgb24, int count );
uint8_t* read_rgb16( const char* filename, int count );
};

                int RGB_to16(int red, int grn, int blu);
		int RGB_24to16(int red, int grn, int blu);
#endif
