// src/main.cc
// This file is part of SlugTerm; see http://chezphil.org/slugterm
// (C) 2005 Philip Endecott
// Modifications for 4D Systems Display Graphics Modules (C) 2008 Sven Killig <sven@killig.de>

// This is version $Name:  $
//   (if there is no version (e.g. V0-1) mentioned in the previous line,
//    this is probably a snapshot from between "official" releases.)

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.


#include "SlugTerm.hh"
#include "UkKeyboardLayout.hh"

#include <unistd.h>
#include <stdio.h>

#include "Exception.hh"

#include <vector>

#include <boost/program_options.hpp>
namespace po = boost::program_options;

using namespace std;


struct options {
  string event_dev_fn;
  int pid;
  string init;
  string fpath;
  int fwidth, fheight;
  SlugTerm::commands_t commands;
};


static int hex_str_to_int(string s)
{
  int i;
  int rc = sscanf(s.c_str(),"%x",&i);
  if (rc!=1) {
    throw "hex_str_to_int failed";
  }
  return i;
}


static struct options parse_command_line(int argc, char* argv[])
{
  po::options_description desc("Allowed options");
  desc.add_options()
    ("help", "show help message")
    ("keyboard,k", po::value<string>(), "keyboard event source")
    ("pid,p",  po::value<string>(), "USB product ID [4]")
    ("init,i",  po::value<string>(), "initialization sequence [58]")
    ("fpath,f",  po::value<string>(), "font")
    ("fwidth,w",  po::value<string>(), "font width for scalable fonts")
    ("fheight,h",  po::value<string>(), "font height for scalable fonts")
    ("commands,c", po::value<SlugTerm::commands_t>()->composing(),
     "command(s) to run in terminal")
    ;

  po::positional_options_description p;
  p.add("keyboard", 1);
  p.add("command", -1);

  po::variables_map vm;
  po::store(po::command_line_parser(argc, argv)
	    .options(desc)
	    .positional(p)
	    .run(),
	    vm);
  po::notify(vm);

  if (vm.count("help") || argc<11) {
    cout << desc << "\n";
    exit(1);
  }

  struct options opts;

  if (!vm.count("keyboard") || !vm.count("pid") || !vm.count("init") || !vm.count("fpath")) {
    throw "option missing";
  }
  opts.event_dev_fn = vm["keyboard"].as<string>();
  opts.pid = hex_str_to_int(vm["pid"].as<string>());
  opts.init = vm["init"].as<string>();
  opts.fpath = vm["fpath"].as<string>();
  if(vm.count("fwidth")) opts.fwidth = atoi( (vm["fwidth"].as<string>()).c_str() );
  if(vm.count("fheight")) opts.fheight = atoi( (vm["fheight"].as<string>()).c_str() );
  opts.commands = vm["commands"].as<SlugTerm::commands_t>();

  return opts;
}


int main(int argc, char* argv[])
{
//  try { try {
//fprintf(stderr,"%i\n", argc);

      UkKeyboardLayout uk_keyboard_layout;

      struct options opts = parse_command_line(argc,argv);

      SlugTerm slugterm(opts.event_dev_fn, uk_keyboard_layout,
			opts.pid, opts.init, opts.fpath, opts.fwidth, opts.fheight,
			opts.commands);
      slugterm.run();

/*    } RETHROW_MISC_EXCEPTIONS }
  catch(Exception& e) {
    e.report(cerr);
    exit(1);
  }*/
}
